;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "dashboard" "20260402.436"
  "A startup screen extracted from Spacemacs."
  '((emacs "27.1"))
  :url "https://github.com/emacs-dashboard/dashboard"
  :commit "176d641a55543bda1f0c7506fb954702350c1857"
  :revdesc "176d641a5554"
  :keywords '("startup" "screen" "tools" "dashboard")
  :authors '(("Rakan Al-Hneiti" . "rakan.alhneiti@gmail.com"))
  :maintainers '(("Jen-Chieh" . "jcs090218@gmail.com")
                 ("Ricardo Arredondo" . "ricardo.richo@gmail.com")))
